"""Define a client to interact with IQVIA."""
from __future__ import annotations

import asyncio
import logging
import sys
from collections.abc import Awaitable, Callable
from typing import Any, cast
from urllib.parse import urlparse

import backoff
from aiohttp import ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientError

from .allergens import Allergens
from .asthma import Asthma
from .const import LOGGER
from .disease import Disease
from .errors import InvalidZipError, RequestError

DEFAULT_RETRIES = 4
DEFAULT_TIMEOUT = 3
DEFAULT_USER_AGENT = (
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_4) "
    + "AppleWebKit/537.36 (KHTML, like Gecko) "
    + "Chrome/65.0.3325.181 Safari/537.36"
)


def is_valid_zip_code(zip_code: str) -> bool:
    """Define whether a string ZIP code is valid.

    Args:
        zip_code: A ZIP code

    Returns:
        Whether the ZIP code is valid.
    """
    return len(zip_code) == 5 and zip_code.isdigit()


class Client:
    """Define the client."""

    def __init__(
        self,
        zip_code: str,
        *,
        logger: logging.Logger | None = None,
        request_retries: int = DEFAULT_RETRIES,
        session: ClientSession | None = None,
    ) -> None:
        """Initialize.

        Args:
            zip_code: A ZIP code.
            logger: An option logger.
            request_retries: The number of retries to give a failed request.
            session: An optional aiohttp ClientSession.

        Raises:
            InvalidZipError: Raised on an invalid ZIP code.
        """
        if not is_valid_zip_code(zip_code):
            raise InvalidZipError(f"Invalid ZIP code: {zip_code}")

        self._request_retries = request_retries
        self._session = session
        self.zip_code = zip_code

        if logger:
            self._logger = logger
        else:
            self._logger = LOGGER

        self.async_request = self._wrap_request_method(self._request_retries)

        self.allergens = Allergens(self)
        self.asthma = Asthma(self)
        self.disease = Disease(self)

    async def _async_request(
        self, method: str, url: str, **kwargs: dict[str, Any]
    ) -> dict[str, Any]:
        """Make an API request.

        Args:
            method: An HTTP method.
            url: An API URL.
            **kwargs: Additional kwargs to send with the request.

        Returns:
            An API response payload.
        """
        url_pieces = urlparse(url)
        kwargs.setdefault("headers", {})
        kwargs["headers"]["Content-Type"] = "application/json"
        kwargs["headers"]["Referer"] = f"{url_pieces.scheme}://{url_pieces.netloc}"
        kwargs["headers"]["User-Agent"] = DEFAULT_USER_AGENT

        if use_running_session := self._session and not self._session.closed:
            session = self._session
        else:
            session = ClientSession(timeout=ClientTimeout(total=DEFAULT_TIMEOUT))

        async with session.request(method, f"{url}/{self.zip_code}", **kwargs) as resp:
            resp.raise_for_status()
            data = await resp.json()

        if not use_running_session:
            await session.close()

        self._logger.debug("Received data for %s: %s", url, data)

        return cast(dict[str, Any], data)

    @staticmethod
    def _handle_on_giveup(_: dict[str, Any]) -> None:
        """Wrap a giveup exception as a RequestError.

        Raises:
            RequestError: Raised upon a request that has failed with multiple tries.
        """
        err_info = sys.exc_info()
        err = err_info[1].with_traceback(err_info[2])  # type: ignore[union-attr]
        raise RequestError(err) from err

    def _wrap_request_method(
        self, request_retries: int
    ) -> Callable[..., Awaitable[dict[str, Any]]]:
        """Wrap the request method in backoff/retry logic.

        Args:
            request_retries: The number of retries to give a failed request.

        Returns:
            A version of the request callable that can do retries.
        """
        return cast(
            Callable[..., Awaitable[dict[str, Any]]],
            backoff.on_exception(
                backoff.expo,
                (asyncio.TimeoutError, ClientError),
                logger=self._logger,
                max_tries=request_retries,
                on_giveup=self._handle_on_giveup,  # type: ignore[arg-type]
            )(self._async_request),
        )

    def disable_request_retries(self) -> None:
        """Disable the request retry mechanism."""
        self.async_request = self._wrap_request_method(1)

    def enable_request_retries(self) -> None:
        """Enable the request retry mechanism."""
        self.async_request = self._wrap_request_method(self._request_retries)



"""Define tests for the "Asthma" object."""
from __future__ import annotations

import json

import aiohttp
import pytest
from aresponses import ResponsesMockServer

from pyiqvia import Client

from .common import TEST_ZIP, load_fixture


@pytest.mark.asyncio
async def test_current(aresponses: ResponsesMockServer) -> None:
    """Test getting current asthma.

    Args:
        aresponses: An aresponses server.
    """
    aresponses.add(
        "www.asthmaforecast.com",
        f"/api/forecast/current/asthma/{TEST_ZIP}",
        "get",
        response=aiohttp.web_response.json_response(
            json.loads(load_fixture("asthma_current_response.json")), status=200
        ),
    )

    async with aiohttp.ClientSession() as session:
        client = Client(TEST_ZIP, session=session)
        current = await client.asthma.current()
        assert len(current["Location"]["periods"]) == 3

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_extended(aresponses: ResponsesMockServer) -> None:
    """Test getting extended asthma info.

    Args:
        aresponses: An aresponses server.
    """
    aresponses.add(
        "www.asthmaforecast.com",
        f"/api/forecast/extended/asthma/{TEST_ZIP}",
        "get",
        response=aiohttp.web_response.json_response(
            json.loads(load_fixture("asthma_extended_response.json")), status=200
        ),
    )

    async with aiohttp.ClientSession() as session:
        client = Client(TEST_ZIP, session=session)
        extended = await client.asthma.extended()
        assert len(extended["Location"]["periods"]) == 5

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_endpoints(aresponses: ResponsesMockServer) -> None:
    """Test getting historic asthma info.

    Args:
        aresponses: An aresponses server.
    """
    aresponses.add(
        "www.asthmaforecast.com",
        f"/api/forecast/historic/asthma/{TEST_ZIP}",
        "get",
        response=aiohttp.web_response.json_response(
            json.loads(load_fixture("asthma_historic_response.json")), status=200
        ),
    )

    async with aiohttp.ClientSession() as session:
        client = Client(TEST_ZIP, session=session)
        historic = await client.asthma.historic()
        assert len(historic["Location"]["periods"]) == 30

    aresponses.assert_plan_strictly_followed()



"""Define common test utilities."""
import os

TEST_BAD_ZIP = "abcde"
TEST_ZIP = "00123"


def load_fixture(filename: str) -> str:
    """Load a fixture.

    Args:
        filename: The filename of the fixtures file to load.

    Returns:
        A string containing the contents of the file.
    """
    path = os.path.join(os.path.dirname(__file__), "fixtures", filename)
    with open(path, encoding="utf-8") as fptr:
        return fptr.read()



"""Define tests for the client object."""
from __future__ import annotations

import asyncio
import json
import logging
from unittest.mock import Mock, patch

import aiohttp
import pytest
from aresponses import ResponsesMockServer

from pyiqvia import Client
from pyiqvia.errors import InvalidZipError, RequestError

from .common import TEST_BAD_ZIP, TEST_ZIP, load_fixture


@pytest.mark.asyncio
async def test_bad_zip() -> None:
    """Test creating a client with a bad ZIP code."""
    with pytest.raises(InvalidZipError):
        async with aiohttp.ClientSession() as session:
            _ = Client(TEST_BAD_ZIP, session=session)


@pytest.mark.asyncio
async def test_create() -> None:
    """Test the creation of a client."""
    async with aiohttp.ClientSession() as session:
        client = Client(TEST_ZIP, session=session)
        assert client.zip_code == TEST_ZIP


@pytest.mark.asyncio
async def test_custom_logger(aresponses: ResponsesMockServer, caplog: Mock) -> None:
    """Test that a custom logger is used when provided to the client.

    Args:
        aresponses: An aresponses server.                            :
        caplog: A mock logging utility.
    """
    caplog.set_level(logging.DEBUG)
    custom_logger = logging.getLogger("custom")

    aresponses.add(
        "www.pollen.com",
        f"/api/forecast/current/pollen/{TEST_ZIP}",
        "get",
        response=aiohttp.web_response.json_response(
            json.loads(load_fixture("allergens_current_response.json")), status=200
        ),
    )

    async with aiohttp.ClientSession() as session:
        client = Client(TEST_ZIP, session=session, logger=custom_logger)
        await client.allergens.current()
        assert any(
            record.name == "custom" and "Received data" in record.message
            for record in caplog.records
        )

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_http_error(aresponses: ResponsesMockServer) -> None:
    """Test an HTTP error.

    Args:
        aresponses: An aresponses server.
    """
    aresponses.add(
        "www.pollen.com",
        f"/api/forecast/outlook/{TEST_ZIP}",
        "get",
        aresponses.Response(text="", status=500),
    )

    with pytest.raises(RequestError):
        async with aiohttp.ClientSession() as session:
            client = Client(TEST_ZIP, session=session, request_retries=1)
            await client.allergens.outlook()

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_request_retries(aresponses: ResponsesMockServer) -> None:
    """Test the request retry logic.

    Args:
        aresponses: An aresponses server.
    """
    aresponses.add(
        "www.pollen.com",
        f"/api/forecast/outlook/{TEST_ZIP}",
        "get",
        aresponses.Response(text="", status=500),
    )
    aresponses.add(
        "www.pollen.com",
        f"/api/forecast/outlook/{TEST_ZIP}",
        "get",
        aresponses.Response(text="", status=500),
    )
    aresponses.add(
        "www.pollen.com",
        f"/api/forecast/outlook/{TEST_ZIP}",
        "get",
        response=aiohttp.web_response.json_response(
            json.loads(load_fixture("allergens_outlook_response.json")), status=200
        ),
    )

    async with aiohttp.ClientSession() as session:
        client = Client(TEST_ZIP, session=session)

        client.disable_request_retries()

        with pytest.raises(RequestError):
            await client.allergens.outlook()

        client.enable_request_retries()

        await client.allergens.outlook()

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_request_timeout() -> None:
    """Test a request timeout."""
    with patch(
        "aiohttp.ClientSession.request", side_effect=asyncio.TimeoutError
    ), pytest.raises(RequestError):
        async with aiohttp.ClientSession() as session:
            client = Client(TEST_ZIP, session=session, request_retries=1)
            await client.allergens.outlook()



"""Define an object to work with asthma endpoints."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .client import Client


class Asthma:
    """Define the "Asthma" object."""

    def __init__(self, client: Client) -> None:
        """Initialize.

        Args:
            client: The Client object.
        """
        self._client = client

    async def current(self) -> dict[str, Any]:
        """Get current asthma info.

        Returns:
            An API response payload.
        """
        return await self._client.async_request(
            "get", "https://www.asthmaforecast.com/api/forecast/current/asthma"
        )

    async def extended(self) -> dict[str, Any]:
        """Get extended asthma info.

        Returns:
            An API response payload.
        """
        return await self._client.async_request(
            "get", "https://www.asthmaforecast.com/api/forecast/extended/asthma"
        )

    async def historic(self) -> dict[str, Any]:
        """Get historic asthma info.

        Returns:
            An API response payload.
        """
        return await self._client.async_request(
            "get", "https://www.asthmaforecast.com/api/forecast/historic/asthma"
        )



"""Run an example script to quickly test."""
import asyncio

from aiohttp import ClientSession

from pyiqvia import Client
from pyiqvia.errors import IQVIAError


async def main() -> None:
    """Run."""
    async with ClientSession() as session:
        try:
            client = Client("17015", session=session)
            print(f"Client instantiated for ZIP: {client.zip_code}")

            print()
            print("Allergen Data:")
            print(await client.allergens.current())
            print(await client.allergens.extended())
            print(await client.allergens.historic())
            print(await client.allergens.outlook())

            print()
            print("Disease Data:")
            print(await client.disease.current())
            print(await client.disease.extended())
            print(await client.disease.historic())

            print()
            print("Asthma Data:")
            print(await client.asthma.current())
            print(await client.asthma.extended())
            print(await client.asthma.historic())
        except IQVIAError as err:
            print(err)


asyncio.run(main())
